/* Copyright (C) 2002-2008 The SSI Project. All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
   3. Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
*/

#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_esolver_init
 * lis_esolver_create
 * lis_esolver_destroy
 * lis_esolver_work_destroy
 * lis_esolver_set_option
 * lis_esolver_get_option
 * lis_esolve
 ************************************************/

LIS_ESOLVER_EXECUTE extern lis_esolver_execute[] = {
  NULL,
  lis_epi, lis_eii, lis_eaii, lis_ecg, lis_eli, lis_esi
};

LIS_ESOLVER_CHECK_PARAMS extern lis_esolver_check_params[] = {
  NULL,
  lis_epi_check_params, lis_eii_check_params, lis_eaii_check_params, 
  lis_ecg_check_params, lis_eli_check_params, lis_esi_check_params
};

LIS_ESOLVER_MALLOC_WORK extern lis_esolver_malloc_work[] = {
  NULL,
  lis_epi_malloc_work, lis_eii_malloc_work, lis_eaii_malloc_work, 
  lis_ecg_malloc_work, lis_eli_malloc_work, lis_esi_malloc_work
};

#define LIS_ESOLVER_OPTION_LEN	8
#define LIS_ESOLVERS_LEN 6
#define LIS_EPRINT_LEN 4
#define LIS_TRUEFALSE_LEN 2

char *LIS_ESOLVER_OPTNAME[] = {
"-emaxiter", "-etol", "-e", "-ss", "-shift", "-eprint", "-initx_ones", "-ie" 
};

int LIS_ESOLVER_OPTACT[] = {
LIS_EOPTIONS_MAXITER, LIS_EPARAMS_RESID, LIS_EOPTIONS_ESOLVER, 
LIS_EOPTIONS_SUBSPACE, LIS_EPARAMS_SHIFT, LIS_EOPTIONS_OUTPUT,
LIS_EOPTIONS_INITGUESS_ONES, LIS_EOPTIONS_INNER_ESOLVER
};

char *lis_esolver_atoi[] = {"pi", "ii", "aii", "cg", "li", "si"};
char *lis_eprint_atoi[] = {"none", "mem", "out", "all"};
char *lis_etruefalse_atoi[] = {"false", "true"};

char *lis_esolvername[] = {"", "Power Iteration", "Inverse Iteration", "Approximate Inverse Iteration", "Conjugate Gradient", "Lanczos Iteration", "Subspace Iteration"};

char *lis_ereturncode[] = {"LIS_SUCCESS", "LIS_ILL_OPTION", "LIS_BREAKDOWN", "LIS_OUT_OF_MEMORY", "LIS_MAXITER", "LIS_NOT_IMPLEMENTED", "LIS_ERR_FILE_IO"};

LIS_VECTOR lis_esolver_evalue = NULL;
int lis_esolver_evalue_count = 0;

LIS_VECTOR lis_esolver_evector = NULL;
int lis_esolver_evector_count = 0;

LIS_VECTOR lis_esolver_residual_history = NULL;
int lis_esolver_residual_history_count = 0;

#undef __FUNC__
#define __FUNC__ "lis_esolver_init"
int lis_esolver_init(LIS_ESOLVER esolver)
{

	LIS_DEBUG_FUNC_IN;

	esolver->A        = NULL;
	esolver->x        = NULL;
	esolver->residual = NULL;
	esolver->evalue   = NULL;
	esolver->evector  = NULL;

	esolver->worklen   = 0;
	esolver->iter      = 0;
	esolver->resid     = 0;
	esolver->times     = 0;
	esolver->itimes    = 0;
	esolver->ptimes    = 0;
	esolver->p_i_times    = 0;
	esolver->p_c_times    = 0;

	esolver->lshift     = 0;
	esolver->tol       = 0;
	esolver->precision = LIS_PRECISION_DOUBLE;

	esolver->options[LIS_EOPTIONS_ESOLVER]               = LIS_ESOLVER_PI;
	esolver->options[LIS_EOPTIONS_MAXITER]               = 1000;
	esolver->options[LIS_EOPTIONS_SUBSPACE]              = 10;
	esolver->options[LIS_EOPTIONS_OUTPUT]                = LIS_FALSE;
	esolver->options[LIS_EOPTIONS_INITGUESS_ONES]        = LIS_TRUE;
	esolver->options[LIS_EOPTIONS_INNER_ESOLVER]         = 2;

	esolver->params[LIS_EPARAMS_RESID - LIS_EOPTIONS_LEN] = 1.0e-5;
	esolver->params[LIS_EPARAMS_SHIFT - LIS_EOPTIONS_LEN] = 0.0;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_create"
int lis_esolver_create(LIS_ESOLVER *esolver)
{
	LIS_DEBUG_FUNC_IN;

	*esolver = NULL;

	*esolver = (LIS_ESOLVER)lis_malloc( sizeof(struct LIS_ESOLVER_STRUCT),"lis_esolver_create::esolver" );
	if( NULL==*esolver )
	{
		LIS_SETERR_MEM(sizeof(struct LIS_ESOLVER_STRUCT));
		return LIS_OUT_OF_MEMORY;
	}
	lis_esolver_init(*esolver);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_work_destroy"
int lis_esolver_work_destroy(LIS_ESOLVER esolver)
{
	int i;

	LIS_DEBUG_FUNC_IN;

	if( esolver && esolver->work )
	{
		for(i=0;i<esolver->worklen;i++) lis_vector_destroy(esolver->work[i]);
		lis_free(esolver->work);
		esolver->work    = NULL;
		esolver->worklen = 0;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_destroy"
int lis_esolver_destroy(LIS_ESOLVER esolver)
{
	LIS_DEBUG_FUNC_IN;

	if( esolver )
	{
		lis_esolver_work_destroy(esolver);
   	        if( esolver->residual ) 
		  {
		    lis_free(esolver->residual);
		    esolver->residual = NULL;
		  }
	        if( esolver->evalue ) 
		  {
		    lis_free(esolver->evalue);
		    esolver->evalue = NULL;
		  }
   	        if( esolver->evector ) 
		  {
		    lis_free(esolver->evector);
		    esolver->evector = NULL;
		  }
		lis_free(esolver);
		esolver = NULL;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolve"
int lis_esolve(LIS_MATRIX A, LIS_VECTOR x, LIS_SCALAR *evalue0, LIS_ESOLVER esolver)
{
        int			nesolver,niesolver,emaxiter; 
	LIS_SCALAR              *evalue;
	LIS_VECTOR	        *evector;
	LIS_SCALAR	        *residual;
	int			err;
	int                     output;
	int                     ss;
	double                  times;
	double                  gshift;

	LIS_DEBUG_FUNC_IN;

	/* begin parameter check */
	err = lis_matrix_check(A,LIS_MATRIX_CHECK_ALL);

	if( err ) return err;
	if( x==NULL )
	{
		LIS_SETERR(LIS_ERR_ILL_ARG,"vector x is undefined\n");
		return LIS_ERR_ILL_ARG;
	}
	if( A->n!=x->n )
	{
		return LIS_ERR_ILL_ARG;
	}
	if( A->n<=0 )
	{
		LIS_SETERR1(LIS_ERR_ILL_ARG,"Dimension n(=%d) of matrix A is less than 0\n",A->n);
		return LIS_ERR_ILL_ARG;
	}

	nesolver = esolver->options[LIS_EOPTIONS_ESOLVER];
	niesolver = esolver->options[LIS_EOPTIONS_INNER_ESOLVER];
	ss = esolver->options[LIS_EOPTIONS_SUBSPACE];
	emaxiter = esolver->options[LIS_EOPTIONS_MAXITER];
	gshift = esolver->params[LIS_EPARAMS_SHIFT - LIS_EOPTIONS_LEN];
	output = esolver->options[LIS_EOPTIONS_OUTPUT];

	if ( output ) if( A->my_rank==0 ) printf("shift = %e\n", gshift);

	if( nesolver < 1 || nesolver > LIS_ESOLVERS_LEN )
	{
		LIS_SETERR2(LIS_ERR_ILL_ARG,"Parameter LIS_EOPTIONS_ESOLVER is %d (Set between 1 to %d)\n",nesolver, LIS_ESOLVERS_LEN);
		return LIS_ERR_ILL_ARG;
	}

	if( niesolver < 1 || niesolver > 4) 
	{
		LIS_SETERR1(LIS_ERR_ILL_ARG,"Parameter LIS_EOPTIONS_INNER_ESOLVER is %d (Set between 1 to 4)\n", niesolver);
		return LIS_ERR_ILL_ARG;
	}

	if ( esolver->options[LIS_EOPTIONS_ESOLVER] == LIS_ESOLVER_SI && niesolver == 4 )
	{
		LIS_SETERR1(LIS_ERR_ILL_ARG,"Parameter LIS_EOPTIONS_INNER_ESOLVER is %d (Set between 1 to 3 for Subspace Iteration)\n", niesolver);
		return LIS_ERR_ILL_ARG;
	}

	if ( esolver->options[LIS_EOPTIONS_ESOLVER] == LIS_ESOLVER_LI && niesolver == 1 )
	{
		LIS_SETERR1(LIS_ERR_ILL_ARG,"Parameter LIS_EOPTIONS_INNER_ESOLVER is %d (Set between 2 to 4 for Lanczos Iteration)\n", niesolver);
		return LIS_ERR_ILL_ARG;
	}

	/* create eigenvalue array */
	if( esolver->evalue ) lis_free(esolver->evalue);
	evalue = (LIS_SCALAR *)lis_malloc((ss+2)*sizeof(LIS_SCALAR),"lis_esolve::evalue");
	if( evalue==NULL )
	{
		LIS_SETERR_MEM((ss+2)*sizeof(LIS_SCALAR));
		esolver->retcode = err;
		return err;
	}
	evalue[0] = 1.0;
	evalue[ss-1] = 1.0;

	/* set initial vector */ 
        if( esolver->options[LIS_EOPTIONS_INITGUESS_ONES] )
	  {
	    if ( output ) lis_printf(A->comm,"Initial vector x = 1\n");
	    lis_vector_set_all(1.0,x);
	  }
        else
	  {
	    if ( output ) lis_printf(A->comm,"Initial vector x = user defined\n");
	  }

	/* create eigenvector array */
	if( esolver->evector ) lis_free(esolver->evector);
	evector = (LIS_VECTOR *)lis_malloc((ss+2)*sizeof(LIS_VECTOR),"lis_esolve::evector");
	if( evector==NULL )
	{
		LIS_SETERR_MEM((ss+2)*sizeof(LIS_VECTOR));
		esolver->retcode = err;
		return err;
	}

	/* create residual history vector */
	if( esolver->residual ) lis_free(esolver->residual);
	residual = (LIS_SCALAR *)lis_malloc((emaxiter+2)*sizeof(LIS_SCALAR),"lis_esolve::residual");
	if( residual==NULL )
	{
		LIS_SETERR_MEM((emaxiter+2)*sizeof(LIS_SCALAR));
		esolver->retcode = err;
		return err;
	}

	esolver->A        = A;
	esolver->evalue   = evalue;
	esolver->x        = x;
	esolver->evector  = evector;
	residual[0] = 1.0;
	esolver->residual = residual;

        if( A->my_rank==0 )
	  {
	    if ( output ) printf("ESOLVER    : %s %d\n\n", lis_esolvername[nesolver],nesolver);
	  }

	times = lis_wtime();

	esolver->ptimes = 0;
	esolver->itimes = 0;
	esolver->p_c_times = 0;
	esolver->p_i_times = 0;

	if (gshift != 0.0) lis_matrix_shift_diagonal(A, gshift);

	/* create work vector */
	err = lis_esolver_malloc_work[nesolver](esolver);
	if( err )
	{
		esolver->retcode = err;
		return err;
	}

	/* execute esolver */
	err = lis_esolver_execute[nesolver](esolver);
	esolver->retcode = err;

	*evalue0 = esolver->evalue[0];
	x = esolver->evector[0];

	esolver->times = lis_wtime() - times; 

	lis_matrix_shift_diagonal(A, -gshift);

        if( A->my_rank==0 )
        {
                if( err )
                {
                  if ( output ) printf("lis_esolve : %s(code=%d)\n\n",lis_ereturncode[err],err);

                }
                else
                {
                  if ( output ) printf("lis_esolve : normal end\n\n");
                }
        }

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_set_optionC"
int lis_esolver_set_optionC(LIS_ESOLVER esolver)
{
	LIS_ARGS	p;

	LIS_DEBUG_FUNC_IN;

	p = cmd_args->next;
	while( p!=cmd_args )
	{
		lis_esolver_set_option2(p->arg1,p->arg2,esolver);
		p = p->next;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_set_option"
int lis_esolver_set_option(char *text, LIS_ESOLVER esolver)
{
	LIS_ARGS	args,p;

	LIS_DEBUG_FUNC_IN;

	lis_text2args(text,&args);
	p = args->next;
	while( p!=args )
	{
		lis_esolver_set_option2(p->arg1,p->arg2,esolver);
		p = p->next;
	}
	lis_args_free(args);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_set_option2"
int lis_esolver_set_option2(char* arg1, char *arg2, LIS_ESOLVER esolver)
{
	int i;

	LIS_DEBUG_FUNC_IN;
	
	for(i=0;i<LIS_ESOLVER_OPTION_LEN;i++)
	{
		if( strcmp(arg1, LIS_ESOLVER_OPTNAME[i])==0 )
		{
			switch( LIS_ESOLVER_OPTACT[i] )
			{
			case LIS_EOPTIONS_ESOLVER:
			  lis_esolver_set_option_esolver(arg2,esolver);
			  break;
                        case LIS_EOPTIONS_OUTPUT:
			  lis_esolver_set_option_print(arg2,esolver);
			  break;
                        case LIS_EOPTIONS_INITGUESS_ONES:
			  lis_esolver_set_option_truefalse(arg2,LIS_EOPTIONS_INITGUESS_ONES,esolver);
			  break;
			case LIS_EOPTIONS_INNER_ESOLVER:
			  lis_esolver_set_option_iesolver(arg2,esolver);
			  break;
			default:
			  if( LIS_ESOLVER_OPTACT[i] < LIS_EOPTIONS_LEN )
			    {
			      sscanf(arg2, "%d", &esolver->options[LIS_ESOLVER_OPTACT[i]]);
			    }
			  else
			    {
			      sscanf(arg2, "%lg", &esolver->params[LIS_ESOLVER_OPTACT[i]-LIS_EOPTIONS_LEN]);
			    }
			  break;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_set_option_esolver"
int lis_esolver_set_option_esolver(char *argv, LIS_ESOLVER esolver)
{
	int  i;

	LIS_DEBUG_FUNC_IN;

	if( argv[0]>='0' && argv[0]<='7' )
	{
		sscanf(argv, "%d", &esolver->options[LIS_EOPTIONS_ESOLVER]);
	}
	else
	{
		for(i=0;i<LIS_ESOLVER_LEN;i++)
		{
			if( strcmp(argv,lis_esolver_atoi[i])==0 )
			{
				esolver->options[LIS_EOPTIONS_ESOLVER] = i+1;
				break;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_set_option_iesolver"
int lis_esolver_set_option_iesolver(char *argv, LIS_ESOLVER esolver)
{
	int  i;

	LIS_DEBUG_FUNC_IN;

	if( argv[0]>='0' && argv[0]<='6' )
	{
		sscanf(argv, "%d", &esolver->options[LIS_EOPTIONS_INNER_ESOLVER]);
	}
	else
	{
		for(i=0;i<LIS_ESOLVER_LEN;i++)
		{
			if( strcmp(argv,lis_esolver_atoi[i])==0 )
			{
				esolver->options[LIS_EOPTIONS_INNER_ESOLVER] = i+1;
				break;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_set_option_print"
int lis_esolver_set_option_print(char *argv, LIS_ESOLVER esolver)
{
	int  i;

	LIS_DEBUG_FUNC_IN;

	if( argv[0]>='0' && argv[0]<='3' )
	{
		sscanf(argv, "%d", &esolver->options[LIS_EOPTIONS_OUTPUT]);
	}
	else
	{
		for(i=0;i<LIS_EPRINT_LEN;i++)
		{
			if( strcmp(argv,lis_eprint_atoi[i])==0 )
			{
				esolver->options[LIS_EOPTIONS_OUTPUT] = i;
				break;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option_truefalse"
int lis_esolver_set_option_truefalse(char *argv, int opt, LIS_ESOLVER esolver)
{
        int  i;

        LIS_DEBUG_FUNC_IN;

        if( argv[0]>='0' && argv[0]<='1' )
        {
                sscanf(argv, "%d", &esolver->options[opt]);
        }
        else
        {
                for(i=0;i<LIS_TRUEFALSE_LEN;i++)
                {
                        if( strcmp(argv,lis_etruefalse_atoi[i])==0 )
                        {
                                esolver->options[opt] = i;
                                break;
                        }
                }
        }
        LIS_DEBUG_FUNC_OUT;
        return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_get_iters"
int lis_esolver_get_iters(LIS_ESOLVER esolver, int *iters)
{
	LIS_DEBUG_FUNC_IN;

	*iters = esolver->iter;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_get_itersex"
int lis_esolver_get_itersex(LIS_ESOLVER esolver, int *iters, int *iters_double, int *iters_quad)
{
	LIS_DEBUG_FUNC_IN;

	*iters = esolver->iter;
	*iters_double = esolver->iter2;
	*iters_quad = esolver->iter - esolver->iter2;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_get_time"
int lis_esolver_get_time(LIS_ESOLVER esolver, double *times)
{
	LIS_DEBUG_FUNC_IN;

	*times  = esolver->times;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_get_timeex"
int lis_esolver_get_timeex(LIS_ESOLVER esolver, double *times, double *itimes, double *ptimes, double *p_c_times, double *p_i_times)
{
	LIS_DEBUG_FUNC_IN;

	*times  = esolver->times;
	if( itimes ) *itimes = esolver->itimes;
	if( ptimes ) *ptimes = esolver->ptimes;
	if( p_c_times ) *p_c_times = esolver->p_c_times;
	if( p_i_times ) *p_i_times = esolver->p_i_times;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_get_residualnorm"
int lis_esolver_get_residualnorm(LIS_ESOLVER esolver, LIS_REAL *residual)
{
	LIS_DEBUG_FUNC_IN;

	*residual  = esolver->resid;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_get_esolver"
int lis_esolver_get_esolver(LIS_ESOLVER esolver, int *nesol)
{
	LIS_DEBUG_FUNC_IN;

	*nesol = esolver->options[LIS_EOPTIONS_ESOLVER];

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_get_evalues"
int lis_esolver_get_evalues(LIS_ESOLVER esolver, LIS_VECTOR v)
{
	int		i,n;
	int             lis_esolver_evalue_count;

	LIS_DEBUG_FUNC_IN;

	lis_esolver_evalue_count = esolver->options[LIS_EOPTIONS_SUBSPACE];

	n = lis_esolver_evalue_count;
	for(i=0;i<n;i++)
	{
		v->value[i] = esolver->evalue[i];
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_get_evectors"
int lis_esolver_get_evectors(LIS_ESOLVER esolver, LIS_VECTOR *evector)
{
	LIS_DEBUG_FUNC_IN;

	evector = esolver->evector;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_esolver_get_rhistory"
int lis_esolver_get_rhistory(LIS_ESOLVER esolver, LIS_VECTOR v)
{
	int		i,n,maxiter;

	maxiter = esolver->iter+1;
        if( esolver->retcode!=LIS_SUCCESS )
	  {
	    maxiter--;
	  }
	n = _min(v->n,maxiter);
	for(i=0;i<n;i++)
	{
		v->value[i] = esolver->residual[i];
	}
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_get_esolvername"
int lis_get_esolvername(int esolver, char *esolvername)
{
	LIS_DEBUG_FUNC_IN;

	if( esolver < 1 || esolver > LIS_ESOLVERS_LEN )
	{
		esolvername = NULL;
		return LIS_FAILS;
	}
	strcpy(esolvername,lis_esolvername[esolver]);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

